import os
import cv2
import numpy as np
import imutils
import math
import asyncio

class RectDetector(object):
    ''' Class for rectangle detection from the mask. '''

    def __init__(self, coef_approx = 0.00001, max_count_step_approx = 300, target_points = 11):
        self.COEF_APPROX = coef_approx
        self.MAX_COUNT_STEP_APPROX = max_count_step_approx
        self.TARGET_POINTS = target_points

    def getApprox(self, c, peri):
        '''function get approx'''
        i = 0
        coef = self.COEF_APPROX
        min_c = 0
        max_c = coef*2
        while i < self.MAX_COUNT_STEP_APPROX:
            approx = cv2.approxPolyDP(c, coef * peri, True)
            cnt = len(approx)
            if cnt == self.TARGET_POINTS:
                break
            if cnt > self.TARGET_POINTS:
                min_c = coef
                max_c = max_c + coef
            if cnt < self.TARGET_POINTS:
                max_c = coef
            coef = (max_c + min_c)/2
            i += 1
        return approx

    def detectRect(self, image):
        '''function detect rect'''

        # load the image and resize it to a smaller factor so that the shapes can be approximated better
        resized = imutils.resize(image, width=1200)
        ratio = image.shape[0] / float(resized.shape[0])

        # convert the resized image to grayscale, blur it slightly, and threshold it
        gray = cv2.cvtColor(resized, cv2.COLOR_BGR2GRAY)
        blurred = cv2.GaussianBlur(gray, (7, 7), 0)
        thresh = cv2.threshold(gray, 60, 255, cv2.THRESH_BINARY)[1]

        # find contours in the thresholded image and initialize the shape detector
        if cv2.__version__[0] == "4":
            cnts, hierarchy = cv2.findContours(thresh.copy(), 1, 2)
        else:
            cnts = cv2.findContours(thresh.copy(), 1, 2)
            cnts = cnts[0] if imutils.is_cv2() else cnts[1]
        res = []
        rows, cols = image.shape[:2]

        # loop over the contours
        for c in cnts:
            # compute the center of the contour, then detect the name of the shape using only the contour
            peri = cv2.arcLength(c, True)
            approx = self.getApprox(c,  peri)
            c = approx.astype("float")

            # multiply the contour (x, y)-coordinates by the resize ratio, then draw the contours and the name of the shape on the image
            c = c.astype("float")
            c *= ratio
            c = c.astype("int")
            res.append(c)
        return res

    def linearLineMatrix(self, p0 ,p1):
        ''' The calculation of the coefficients of the matrix describing the line by two points. '''
        x1 = float(p0[0])
        y1 = float(p0[1])

        x2 = float(p1[0])
        y2 = float(p1[1])

        A = y1 - y2
        B = x2 - x1
        C = x2*y1-x1*y2
        return [A, B, C]

    def fline(self, p0, p1):
        ''' Calculation of the angle of inclination of a straight line by 2 points. '''
        x1 = float(p0[0])
        y1 = float(p0[1])

        x2 = float(p1[0])
        y2 = float(p1[1])

        if (x1 - x2 == 0):
            k = 100000000
            b = y2
        else:
            k = (y1 - y2) / (x1 - x2)
            b = y2 - k*x2
        r = math.atan(k)
        a = math.degrees(r)
        a180 = a
        if (a < 0 ):
            a180 = 180 + a
        return [k, b, a, a180, r]

    def distance(self, p0, p1):
        return math.sqrt((p0[0] - p1[0])*(p0[0] - p1[0]) + (p0[1] - p1[1])*(p0[1] - p1[1]))

    def rotate_points(self,rect):
        rect[0], rect[1], rect[2], rect[3] = rect[1], rect[2], rect[3], rect[0]
        return rect

    def findWidth(self, rect):
        line1 = self.distance(rect[0], rect[1])
        line3 = self.distance(rect[2], rect[3])
        h = np.mean([line1, line3])
        return h

    def findHieght(self, rect):
        line2 = self.distance(rect[1], rect[2])
        line4 = self.distance(rect[3], rect[0])
        w = np.mean([line2, line4])
        return w

    def to_pretty_point(self, points):
        sortedOnX = sorted([point for point in points], key=lambda x: x[0] )
        res = [[0, 0], [0, 0], [0, 0], [0, 0]]
        if sortedOnX[0][1] < sortedOnX[1][1]:
            res[0], res[3] = sortedOnX[0], sortedOnX[1]
        else:
            res[0], res[3] = sortedOnX[1], sortedOnX[0]

        if sortedOnX[2][1] < sortedOnX[3][1]:
            res[1], res[2] = sortedOnX[2], sortedOnX[3]
        else:
            res[1], res[2] = sortedOnX[3], sortedOnX[2]
        return res


    def rotate_to_pretty(self, points):
        new_arr = []
        points = self.to_pretty_point(points)
        new_arr.append(points)
        w = self.findWidth(points)
        h = self.findHieght(points)

        if h > w:
            h, w = w, h
            points = self.reshapePoints(points, 1)
        return points, w, h

    def detetct_pretty_w_h_to_zone(self, w, h, coef=4.6):
        return int(h*coef), int(h)

    async def get_cv_zoneRGB_async(self, img, rect, gw = 0, gh = 0, coef=4.6):
        rect, w, h = self.rotate_to_pretty(rect)
        if (gw == 0 or gh == 0):
            w, h = self.detetct_pretty_w_h_to_zone(w, h, coef)
        else:
            w, h = gw, gh
        pts1 = np.float32(rect)
        pts2 = np.float32(np.array([[0, 0], [w, 0], [w, h], [0, h]]))
        M = cv2.getPerspectiveTransform(pts1, pts2)
        dst = cv2.warpPerspective(img,M,(w,h))
        return dst;

    async def get_cv_zonesRGB_async(self, img, rects, gw = 0, gh = 0, coef=4.6):
        loop = asyncio.get_event_loop()
        promises = [loop.create_task(self.get_cv_zoneRGB_async(img, rect, gw = gw, gh = gh, coef=coef)) for rect in rects]
        if bool(promises):
            await asyncio.wait(promises)
        return [promise.result() for promise in promises]

    async def get_cv_zonesBGR_async(self, img, rects, gw = 0, gh = 0, coef=4.6):
        loop = asyncio.get_event_loop()
        promises = [loop.create_task(self.get_cv_zoneRGB_async(img, rect, gw = gw, gh = gh, coef=coef)) for rect in rects]
        if bool(promises):
            await asyncio.wait(promises)
        return [cv2.cvtColor(promise.result(), cv2.COLOR_RGB2BGR) for promise in promises]

    def get_cv_zonesRGB(self, img, rects, gw = 0, gh = 0, coef=4.6):
        dsts = []
        for rect in rects:
            rect, w, h = self.rotate_to_pretty(rect)
            if (gw == 0 or gh == 0):
                w, h = self.detetct_pretty_w_h_to_zone(w, h, coef)
            else:
                w, h = gw, gh
            pts1 = np.float32(rect)
            pts2 = np.float32(np.array([[0, 0], [w, 0], [w, h], [0, h]]))
            M = cv2.getPerspectiveTransform(pts1, pts2)
            dst = cv2.warpPerspective(img,M,(w,h))
            dsts.append(dst)
        return dsts;

    def get_cv_zonesBGR(self, img, rects, gw = 0, gh = 0, coef=4.6):
        dsts = self.get_cv_zonesRGB(img, rects, gw, gh, coef)
        bgrDsts = []
        for dst in dsts:
            dst = cv2.cvtColor(dst, cv2.COLOR_RGB2BGR)
            bgrDsts.append(dst)
        return bgrDsts

    def makeUglyPoints(self, points):
        return [[p] for p in points]

    def findDistances(self, points):
        ''' Getting an array with line characteristics '''
        distanses = []
        cnt = len(points)
        for i in range(cnt):
            p0 = i
            if (i < cnt-1):
                p1 = i+1
            else:
                p1 = 0
            distanses.append({ "d": self.distance(points[p0][0], points[p1][0]), "p0": p0, "p1":p1,
                              "matrix": self.linearLineMatrix(points[p0][0],points[p1][0]),
                              "coef": self.fline(points[p0][0],points[p1][0])})
        return distanses

    def clacRectLines(self, distanses):
        ''' Sort lines to length '''
        return sorted(distanses, key=lambda x: x["d"])

    def filterInterestedLines(self, interestedLines,minElements,thresholdPercentage):
        if (len(interestedLines) > minElements):
            threshold = interestedLines[len(interestedLines)-1]["d"]*thresholdPercentage
            while True:
                interestedLinesFilterd = [x for x in interestedLines if x["d"] >= threshold]
                if len(interestedLinesFilterd) > minElements:
                    break
                threshold = threshold*.9
            # interestedLinesForFilter = interestedLines[:len(interestedLines)-8]
            # interestedLinesFiltered = [x for x in interestedLinesForFilter if x["d"] < threshold]
            return interestedLinesFilterd
        else:
            return interestedLines

        #threshold = interestedLines[len(interestedLines)-1]["d"]*thresholdPercentage
        #return [x for x in interestedLines if x["d"] >= threshold]

    def linearLineMatrixCrossPoint(self, p0,p1,point):
        arr = self.linearLineMatrix(p0,p1)
        x = point[0]
        y = point[1]
        arr[2] = arr[0]*x + arr[1]*y
        return arr

    def detectSimilarLines(self, C,targetLine, significantAlignPercentage=0.3):
        arr = []
        for i in range(len(C)):
            if ((targetLine['p0'] == C[i]['p1']) or (targetLine['p1'] == C[i]['p0'])):
                arr.append(C[i])
        if len(arr) > 1:
            dProp = arr[0]['d']/arr[1]['d'];
            if (dProp <= (1-significantAlignPercentage)):
                arr = [arr[1]]
            else:
                if (dProp >= (1-significantAlignPercentage)):
                    arr = [arr[0]]
                else:
                    arr = []
        return arr

    def detectAlignmentLines(self, targetLines,A,B):
        arrA0 = self.detectSimilarLines(A,targetLines[0])
        arrB0 = self.detectSimilarLines(B,targetLines[1])
        arrA1 = self.detectSimilarLines(A,targetLines[2])
        arrB1 = self.detectSimilarLines(B,targetLines[3])
        return np.array([arrA0,arrB0,arrA1,arrB1])

    def makeAligmentKeyPoints(self, line, arr):
        commonPoint = -1
        startPoint = -1
        endPoint = -1
        subline = arr[0]
        if subline['p0'] == line['p1']:
            commonPoint = line['p1']
            startPoint = line['p0']
            endPoint = subline['p1']
        else:
            commonPoint = line['p0']
            startPoint = line['p1']
            endPoint = subline['p0']
        return { "cPoint": commonPoint, "sPoint": startPoint, "ePoint": endPoint }

    def makeAligmentMatrix(self, line, arr, points):
        if (len(arr) == 0):
            return line["matrix"]
        else:
            pObj = self.makeAligmentKeyPoints(line, arr)
            return self.linearLineMatrixCrossPoint(points[pObj["sPoint"]][0],points[pObj["ePoint"]][0],points[pObj["cPoint"]][0])

    def makeTargetPoints2(self, targetLines,targetAlignmentLines,points):

        line0matrix = self.makeAligmentMatrix(targetLines[0],targetAlignmentLines[0],points)
        line1matrix = self.makeAligmentMatrix(targetLines[1],targetAlignmentLines[1],points)
        line2matrix = self.makeAligmentMatrix(targetLines[2],targetAlignmentLines[2],points)
        line3matrix = self.makeAligmentMatrix(targetLines[3],targetAlignmentLines[3],points)

        point1 = self.detectIntersection(line0matrix,line1matrix)
        point2 = self.detectIntersection(line1matrix,line2matrix)
        point3 = self.detectIntersection(line2matrix,line3matrix)
        point4 = self.detectIntersection(line3matrix,line0matrix)
        return np.array([point1,point2,point3,point4])

    def gDiff(self, a, b):
        d1 = abs(a-b)
        d2 = 180-(d1)
        if (d1<d2):
            return d1
        else:
            return d2

    def cdist(self, X, centroids):
        lines = []
        for x in X:
            line = []
            for c in centroids:
                line.append(self.gDiff(x,c))
            lines.append(line)
        return np.array(lines)

    def cmean(self, X, centroid):
        lines = []
        for i in range(0,len(X)):
            x = X[i]
            if abs(x-centroid)>abs(x-180-centroid):
                X[i] = x-180
        return np.mean(X, axis=0)

    def gKMeans(self, X, maxDeep=0, currentLevel=0):
        ''' Simplified implementation of k-means clustering for 2 classes over a one-character array '''

        parts = np.int32(np.random.random(2)*currentLevel)
        centroids = np.array([0+parts[0],90+parts[1]])

        # Consider the distance from observations to centroid
        for i in range(0,10):
            distances = self.cdist(X, centroids)

            # look to which centroid of each point is closest
            labels = distances.argmin(axis=1)
            c0 = self.cmean(X[labels == 0], centroids[0])
            c1 = self.cmean(X[labels == 1], centroids[1])
            if not np.isnan(c0):
                centroids[0] = c0
            if not np.isnan(c1):
                centroids[1] = c1
        if (len(X[labels == 0])==0) or (len(X[labels == 1])==0):
            currentLevel = currentLevel+1
            if maxDeep>currentLevel:
                labels,distances,centroids = self.gKMeans(X,maxDeep,currentLevel)
        return [labels,distances,centroids]

    def gKMeansMajorLines(self, interestedLines,minElements=5):
        Xfull = [item["coef"][3] for item in interestedLines]
        Xfull = np.float32(Xfull)

        interestedLinesFiltered = self.filterInterestedLines(interestedLines,4,0.2)
        X = [item["coef"][3] for item in interestedLinesFiltered]
        X = np.float32(X)

        labels,distances,centroids = self.gKMeans(X,10)
        distances = self.cdist(Xfull, centroids)
        labels = distances.argmin(axis=1)
        if (len(Xfull[labels == 0])==0) or (len(Xfull[labels == 1])==0):
            if (minElements<len(interestedLines)):
                minElements = minElements+1
                labels = self.gKMeansMajorLines(interestedLines,minElements)
        return labels

    def detectIntersection(self, matrix1,matrix2):
        X = np.array([matrix1[:2],matrix2[:2]])
        y = np.array([matrix1[2], matrix2[2]])
        return np.linalg.lstsq(X, y, rcond=None)[0]

    def isHaveCommonPoint(self, dist1, dist2):
        if (dist1["p0"] == dist2["p1"]) or (dist1["p1"] == dist2["p0"]):
            return True
        else:
            return False

    def isHaveCommonPointRecursive(self, C, i1, i2, deep=0):
        deep = deep + 1;
        if len(C[i1]["links"]) < 1 or deep > len(C):
            return False
        for iTarget in C[i1]["links"]:
            if iTarget == i2:
                return True
            else:
                srabotka = self.isHaveCommonPointRecursive(C, iTarget, i2, deep)
                if srabotka:
                    return True
        return False

    def makeCommonPointsLinks(self, C):
        for i in range(0, len(C)):
            p0 = C[i]["p0"]
            p1 = C[i]["p1"]
            C[i]["links"] = []
            for iTarget in range(0, len(C)):
                if i != iTarget:
                    if self.isHaveCommonPoint(C[i], C[iTarget]):
                        C[i]["links"].append(iTarget)


    def detectIndexesParalelLines(self, C):
        self.makeCommonPointsLinks(C)
        pointIdx1 = len(C) - 1
        idx = 2
        pointIdx2 = len(C) - idx

        while self.isHaveCommonPointRecursive(C,pointIdx1,pointIdx2):
            #pointIdx1 = len(C) - idx
            idx = idx + 1
            pointIdx2 = len(C) - idx
        return [pointIdx1, pointIdx2]

    def makeTargetLines(self, A, B):
        pointsA = self.detectIndexesParalelLines(A)
        pointsB = self.detectIndexesParalelLines(B)
        return np.array([A[pointsA[0]], B[pointsB[0]], A[pointsA[1]], B[pointsB[1]]])

    def makeTargetPoints(self, targetLines):
       point1 = self.detectIntersection(targetLines[0]["matrix"],targetLines[1]["matrix"])
       point2 = self.detectIntersection(targetLines[1]["matrix"],targetLines[2]["matrix"])
       point3 = self.detectIntersection(targetLines[2]["matrix"],targetLines[3]["matrix"])
       point4 = self.detectIntersection(targetLines[3]["matrix"],targetLines[0]["matrix"])
       return np.array([point1,point2,point3,point4])

    def reshapePoints(self, targetPoints, startIdx):
        if [startIdx > 0]:
            part1 = targetPoints[:(startIdx)]
            part2 = targetPoints[(startIdx):]
            targetPoints = np.concatenate((part2, part1))
        return targetPoints

    def findMinXIdx(self, targetPoints):
        minXIdx = 3
        for i in range(0,len(targetPoints)):
            if (targetPoints[i][0] < targetPoints[minXIdx][0]):
                minXIdx = i
            if (targetPoints[i][0] == targetPoints[minXIdx][0]) and (targetPoints[i][1] < targetPoints[minXIdx][1]):
                minXIdx = i
        return minXIdx

    def detectIntersectionNormDD(self, matrix1, matrix2, d1, d2):
        X = np.array([matrix1[:2], matrix2[:2]])
        c0 = matrix1[2] - d1 * (matrix1[0]*matrix1[0] + matrix1[1]*matrix1[1]) ** 0.5
        c1 = matrix2[2] - d2 * (matrix2[0]*matrix2[0] + matrix2[1]*matrix2[1]) ** 0.5
        y = np.array([c0, c1])
        return np.linalg.lstsq(X, y, rcond=None)[0]

    def detectIntersectionNormD(self, matrix1, matrix2, d):
        X = np.array([matrix1[:2], matrix2[:2]])
        c0 = matrix1[2] - d * (matrix1[0]*matrix1[0] + matrix1[1]*matrix1[1]) ** 0.5
        c1 = matrix2[2] - d * (matrix2[0]*matrix2[0] + matrix2[1]*matrix2[1]) ** 0.5
        y = np.array([c0, c1])
        return np.linalg.lstsq(X, y, rcond=None)[0]

    def addOffset(self, targetPoints, offsetHorisontal, offsetVertical):
       distanses = self.findDistances(self.makeUglyPoints(targetPoints))
       points=[]
       cnt = len(distanses)
       offsetFlag = distanses[0]['d']>distanses[1]['d']
       for i in range(cnt):
           iPrev = i
           iNext = i+1
           if (iNext == cnt):
               iNext=0
           if offsetFlag:
               offset1 = offsetVertical
               offset2 = offsetHorisontal
           else:
               offset2 = offsetVertical
               offset1 = offsetHorisontal
           offsetFlag = not offsetFlag
           points.append(self.detectIntersectionNormDD(distanses[iPrev]['matrix'],distanses[iNext]['matrix'],offset1,offset2))
       return np.array(points)

    def fixClockwise(self, targetPoints):
        stat1=self.fline(targetPoints[0],targetPoints[1])
        stat2=self.fline(targetPoints[0],targetPoints[2])

        if targetPoints[0][0] == targetPoints[1][0] and (targetPoints[0][1] > targetPoints[1][1]):
            stat1[2] = -stat1[2]

        if (stat2[2] < stat1[2]):
            targetPoints = np.array([targetPoints[0],targetPoints[3],targetPoints[2],targetPoints[1]])

        return targetPoints

    # fix rectangle points
    # http://www.math.by/geometry/eqline.html
    # https://xn--80ahcjeib4ac4d.xn--p1ai/information/solving_systems_of_linear_equations_in_python/
    def detectParalelMatrix(self, matrix,point):
        new_matrix = matrix
        new_matrix[2] = matrix[0]*point[0]+matrix[1]*point[1]
        return new_matrix

    def detectIntersectionParalelLine(self, line1,line2,point):
        line2matrix = self.detectParalelMatrix(line2['matrix'],point)
        return self.detectIntersection(line1['matrix'],line2matrix)

    def detectUnstableLines(self, distanses,points,d0,d1):
        angle0 = distanses[0]["coef"][3]-distanses[1]["coef"][3]
        if d0<d1:
            angle1 = distanses[0]["coef"][3]-distanses[3]["coef"][3]
            if abs(angle0)>abs(angle1):
                if (distanses[2]["d"] < distanses[0]["d"]):
                    points[3] = self.detectIntersectionParalelLine(distanses[2],distanses[1],points[distanses[0]["p0"]])
                else:
                    points[0] = self.detectIntersectionParalelLine(distanses[0],distanses[1],points[distanses[2]["p1"]])
            else:
                if (distanses[2]["d"] <distanses[0]["d"]):
                    points[2] = self.detectIntersectionParalelLine(distanses[2],distanses[3],points[distanses[0]["p1"]])
                else:
                    points[1] = self.detectIntersectionParalelLine(distanses[0],distanses[3],points[distanses[2]["p0"]])
        else:
            angle1 = distanses[1]["coef"][3]-distanses[2]["coef"][3]
            if abs(angle0)>abs(angle1):
                if (distanses[3]["d"] <distanses[1]["d"]):
                    points[3] = self.detectIntersectionParalelLine(distanses[3],distanses[0],points[distanses[1]["p1"]])
                else:
                    points[2] = self.detectIntersectionParalelLine( distanses[1],distanses[0],points[distanses[3]["p0"]])
            else:
                if (distanses[3]["d"] <distanses[1]["d"]):
                    points[0] = self.detectIntersectionParalelLine(distanses[3],distanses[2],points[distanses[1]["p0"]])
                else:
                    points[1] = self.detectIntersectionParalelLine(distanses[1],distanses[2],points[distanses[3]["p1"]])
        return points

    def fixRectangle(self, points, fixRectangleAngle=3):
        distanses = self.findDistances(self.makeUglyPoints(points))
        d0 = self.gDiff(distanses[0]["coef"][3],distanses[2]["coef"][3])
        d1 = self.gDiff(distanses[1]["coef"][3],distanses[3]["coef"][3])
        if (d0>fixRectangleAngle) or (d1>fixRectangleAngle):
            points = self.detectUnstableLines(distanses,points,d0,d1)
        if (d0>fixRectangleAngle) or (d1>fixRectangleAngle):
            points = self.fixRectangle(points, fixRectangleAngle)
        return points

    def findMaxs(self, points):
        maxX = max(points, key=lambda p: p[0])
        maxY = max(points, key=lambda p: p[1])
        minX = min(points, key=lambda p: p[0])
        minY = min(points, key=lambda p: p[1])
        return maxX, maxY, minX, minY

    def uniquePoints(self, arrPoints):
       arrPointsNew = []
       for i in range(0,len(arrPoints)):
           points = arrPoints[i]
           pointsNew = []
           pStart = [[-1,-1]]
           for j in range(0,len(points)):
               p = points[j]
               if (not (pStart==p).all()):
                   pointsNew.append(p)
               pStart = p
           arrPointsNew.append(pointsNew)
       return np.array(arrPointsNew)

    def checkIfIsSquare(self, points, coef = 2.):
        points = np.array(points)

        d1 = self.distance(points[0], points[1])
        d2 = self.distance(points[1], points[2])

        distanses = self.findDistances(self.makeUglyPoints(points))
        angle = distanses[0]["coef"][3]

        if d1 > d2:
            d1, d2 = d2, d1
        #print(d2/d1)
        #print(angle)
        if (d2/d1) <= coef:
            return True
        return False

    def sortBySize(self, mainArr):
        if len(mainArr) < 2:
            return mainArr
        return sorted(mainArr, key=lambda x: cv2.contourArea(np.array(x).astype(int)), reverse=True)

    async def detectOneAsync(self, image, outboundWidthOffset=3, outboundHeightOffset=0, fixRectangleAngle=3, fixGeometry=0):
        arrPoints = self.detectRect(image)
        arrPoints = self.uniquePoints(arrPoints)

        for points in arrPoints:
            distanses = self.findDistances(points)
            interestedLines = self.clacRectLines(distanses)

            # Get 4 lines that are part of a rectangle describing the license plate area.
            labels = self.gKMeansMajorLines(interestedLines)
            A = []
            B = []
            for i in range(len(interestedLines)):
                if (labels[i] == 0):
                    A.append(interestedLines[i])
                else:
                    B.append(interestedLines[i])
            targetLines = self.makeTargetLines(A,B)
            if fixGeometry:
                targetAlignmentLines = self.detectAlignmentLines(targetLines,A,B)
                targetPoints = self.makeTargetPoints2(targetLines,targetAlignmentLines,arrPoints[0])
            else:
                targetPoints = np.float32(self.makeTargetPoints(targetLines))

            minXIdx = self.findMinXIdx(targetPoints)

            targetPoints=self.reshapePoints(targetPoints,minXIdx)
            targetPoints=self.fixClockwise(targetPoints)
            targetPoints = self.fixRectangle(targetPoints, fixRectangleAngle)

            if outboundWidthOffset or outboundHeightOffset:
                targetPoints=self.addOffset(targetPoints,outboundWidthOffset,outboundHeightOffset)

            return targetPoints

    async def detectAsync(self, images, outboundWidthOffset=3, outboundHeightOffset=0, fixRectangleAngle=3, fixGeometry=0):
         ''' Main method '''
         loop = asyncio.get_event_loop()
         promises = [loop.create_task(self.detectOneAsync(image, outboundWidthOffset=outboundWidthOffset, outboundHeightOffset=outboundHeightOffset, fixRectangleAngle=fixRectangleAngle, fixGeometry=fixGeometry)) for image in images]
         if bool(promises):
            await asyncio.wait(promises)
         return np.array([promise.result() for promise in promises])

    def detect(self, images, outboundWidthOffset=3, outboundHeightOffset=0, fixRectangleAngle=3, fixGeometry=1):
        ''' Main method '''
        resPoints = []
        for image in images:
            res = []
            arrPoints = self.detectRect(image)
            arrPoints = self.uniquePoints(arrPoints)

            for points in arrPoints:
                distanses = self.findDistances(points)
                interestedLines = self.clacRectLines(distanses)

                # Get 4 lines that are part of a rectangle describing the license plate area.
                labels = self.gKMeansMajorLines(interestedLines)
                A = []
                B = []
                for i in range(len(interestedLines)):
                    if (labels[i] == 0):
                        A.append(interestedLines[i])
                    else:
                        B.append(interestedLines[i])
                targetLines = self.makeTargetLines(A,B)

                #print("targetLines")
                #print(targetLines)

                if fixGeometry:
                    targetAlignmentLines = self.detectAlignmentLines(targetLines,A,B)
                    targetPoints = self.makeTargetPoints2(targetLines,targetAlignmentLines,arrPoints[0])
                else:
                    targetPoints = np.float32(self.makeTargetPoints(targetLines))

                minXIdx = self.findMinXIdx(targetPoints)

                targetPoints=self.reshapePoints(targetPoints,minXIdx)
                targetPoints=self.fixClockwise(targetPoints)
                targetPoints = self.fixRectangle(targetPoints, fixRectangleAngle)

                if outboundWidthOffset or outboundHeightOffset:
                    targetPoints=self.addOffset(targetPoints,outboundWidthOffset,outboundHeightOffset)

                res.append(targetPoints)

            resPoints.append(res[0])

        return np.array(resPoints)


import os
import json
import cv2
import numpy as np
import random
from keras.utils import to_categorical
from .aug import aug

class ImgGenerator:

    def __init__(self,
                 dirpath,
                 img_w, img_h,
                 batch_size,
                 labels_counts):

        self.HEIGHT = img_h
        self.WEIGHT = img_w
        self.batch_size = batch_size

        self.labels_counts = labels_counts

        img_dirpath = os.path.join(dirpath, 'img')
        ann_dirpath = os.path.join(dirpath, 'ann')
        self.samples = []
        for filename in os.listdir(img_dirpath):
            name, ext = os.path.splitext(filename)
            if ext == '.png':
                img_filepath = os.path.join(img_dirpath, filename)
                json_filepath = os.path.join(ann_dirpath, name + '.json')
                if os.path.exists(json_filepath):
                    description = json.load(open(json_filepath, 'r'))
                    self.samples.append([img_filepath, [description["state_id"], description["region_id"], description["count_lines"]]])

        self.n = len(self.samples)
        self.indexes = list(range(self.n))
        self.cur_index = 0
        self.count_ep = 0
        self.count_ep_need_to_aug = 1

    def build_data(self):
        self.paths = []
        self.discs = []
        for i, (img_filepath, disc) in enumerate(self.samples):
            self.paths.append(img_filepath)
            self.discs.append(
                [
                    to_categorical(disc[0], self.labels_counts[0]),
                    to_categorical(disc[1], self.labels_counts[1]),
                    to_categorical(disc[2], self.labels_counts[2])
                ]
            )

    def normalize(self, img, with_aug=False):
        if with_aug:
            imgs = aug([img])
            img = imgs[0]
        img = cv2.resize(img, (self.WEIGHT, self.HEIGHT))
        img = img.astype(np.float32)

        # advanced normalisation
        img_min = np.amin(img)
        img -= img_min
        img_max = np.amax(img)
        img /= (img_max or 1)
        img[img == 0] = .0001
        return img

    def next_sample(self):
        self.cur_index += 1
        if self.cur_index >= self.n:
            self.count_ep += 1
            self.cur_index = 0
            random.shuffle(self.indexes)
        return self.paths[self.indexes[self.cur_index]], self.discs[self.indexes[self.cur_index]]

    def generator(self):
        while True:
            Ys = [[], [], []]
            Xs = []
            for i in np.arange(self.batch_size):
                x, y = self.next_sample()
                img = cv2.imread(x)
                if self.count_ep >= self.count_ep_need_to_aug:
                    x = self.normalize(img, with_aug=1)
                else:
                    x = self.normalize(img)
                Xs.append(x)
                Ys[1].append(y[0])
                Ys[0].append(y[1])
                Ys[2].append(y[2])
            yield np.array(Xs), Ys



import imgaug as ia
from imgaug import augmenters as iaa
import numpy as np

ia.seed(1)

def aug(imgs):
    sometimes = lambda aug: iaa.Sometimes(0.5, aug)
    sometimes = lambda aug: iaa.Sometimes(0.5, aug)
    seq = iaa.Sequential(
        [
             sometimes(iaa.Crop(percent=(0, 0.01))),
             iaa.Affine(
                 scale={"x": (0.995, 1.01), "y": (0.995, 1.01)},
                 translate_percent={"x": (-0.01, 0.01), "y": (-0.01, 0.01)},
                 rotate=(-3, 3),
                 shear=(-3, 3),
                 #cval=(0, 255)
             ),
              iaa.SomeOf((0, 5),
              [
                      sometimes(iaa.OneOf([
                              iaa.OneOf([
                                  iaa.GaussianBlur((1, 1.2)),
                                  iaa.AverageBlur(k=(1, 3)),
                                  iaa.MedianBlur(k=(1, 3))
                              ]),
                              iaa.ContrastNormalization((0.5, 1.5), per_channel=0.5),
                              iaa.Grayscale(alpha=(0.0, 1.0)),
                              iaa.OneOf([
                                  iaa.EdgeDetect(alpha=(0, 0.7)),
                                  iaa.DirectedEdgeDetect(
                                      alpha=(0, 0.7), direction=(0.0, 1.0)
                                  ),
                              ]),
                      ])),
                      sometimes(iaa.Sharpen(alpha=(0, 1.0), lightness=(0.75, 1.5))),
                      sometimes(iaa.AdditiveGaussianNoise(loc=0, scale=(0.0, 0.005*255), per_channel=0.001)),
                      sometimes(iaa.Dropout((0.001, 0.01), per_channel=0.5)),
                      sometimes(iaa.Add((-10, 10), per_channel=0.5)),
                      sometimes(iaa.Multiply((0.5, 1.5), per_channel=0.5)),
                      sometimes(iaa.ElasticTransformation(alpha=(0.1, 0.2), sigma=0.05)),
                      sometimes(iaa.PiecewiseAffine(scale=(0.001, 0.005)))
                  ],
                  random_order=True
              )
         ],
        random_order=True
    )
    return seq.augment_images(imgs)


import re
import numpy as np
import string

class xx_xx():
    def __init__(self, standart = "", allowed_liters = string.ascii_letters, black_list=["\s", '\*', '\,', '\.', '\-', "\'", '\"', "\’", "_", "\+"]):
        self.STANDART = self.check_pattern_standart(standart)
        self.ALLOWED_LITERS = allowed_liters
        self.BLACK_LIST = black_list
        self.ALLOWED_NUMBERS = [str(item) for item in np.arange(10)]
        self.REPLACEMENT = {
            "#": {
                "I": "1",
                "Z": "2",#7
                "O": "0",
                "Q": "0",
                "B": "8",
                "D": "0",
                "S": "5",#8
                "T": "7"
            },
            "@": {
                "/": "I",
                "|": "I",
                "¥": "X",
                "€": "C"
            }
        }

    def delete_all_black_list_characters(self, text):
        reg = "[{}]".format("".join(self.BLACK_LIST))
        return re.sub(re.compile(reg), "", text).replace("\\", "/").replace("\[", "|").replace("\]", "|") # replace hard

    def check_pattern_standart(self, standart):
        if not re.match(r"^[#@]*$", standart):
            raise Exception("Standart {} not correct".format(standart))
        return standart

    def check_is_str(self, text):
        if type(text) is not str:
            raise ValueError("{} is not str".format(text))
        return text

    def findFully(self, text):
        reg = ""
        for item in self.STANDART:
            if item == "@":
                reg = "{}[{}]".format(reg,"".join(self.ALLOWED_LITERS))
            elif item == "#":
                reg = "{}[{}]".format(reg,"".join(self.ALLOWED_NUMBERS))
        reg_all = re.compile(reg)
        return re.search(reg_all, text)

    def replace(self, text):
        res = ""
        for i in np.arange(len(self.STANDART)):
            if self.STANDART[i] == "#":
                l_dict = self.ALLOWED_NUMBERS
            elif self.STANDART[i] == "@":
                l_dict = self.ALLOWED_LITERS
            if text[i] in l_dict:
                res = "{}{}".format(res, text[i])
            else:
                replace_l = self.REPLACEMENT[self.STANDART[i]][text[i]]
                res = "{}{}".format(res, replace_l)
        return res

    def findSimilary(self, text):
        vcount = len(text) - len(self.STANDART) + 1
        reg = ""
        for item in self.STANDART:
            if item == "@":
                dop = list(self.REPLACEMENT["@"].keys())
                main = self.ALLOWED_LITERS
            elif item == "#":
                dop = list(self.REPLACEMENT["#"].keys())
                main = self.ALLOWED_NUMBERS
            buf_reg = "".join(main + dop)
            reg = "{}[{}]".format(reg, buf_reg)
        reg_sim = re.compile(reg)
        for i in np.arange(vcount):
            buff_text = text[int(i):int(len(self.STANDART)+i)]
            match = re.search(reg_sim, buff_text)
            if match:
                return self.replace(match.group(0))
        return text

    def find(self, text, strong=True):
        text = self.check_is_str(text)
        text = self.delete_all_black_list_characters(text)
        text = text.upper()

        if len(text) < len(self.STANDART):
            return text

        if len(self.STANDART):
            match = self.findFully(text)
            if match :
                return match.group(0)

        if not strong:
            return self.findSimilary(text)
        return text


import tensorflow as tf
from tensorflow.python.framework import graph_io
from tensorflow.python.tools import freeze_graph
from tensorflow.core.protobuf import saver_pb2
from tensorflow.python.training import saver as saver_lib
from tensorflow.python.framework.graph_util import convert_variables_to_constants
from keras import backend as K

def convert_keras_to_freeze_pb(model, frozen_model_path):
        out_names = ",".join([layer.name.split(":")[0]  for layer in model.outputs])
        inp_names = ",".join([layer.name.split(":")[0]  for layer in model.inputs])
        print("OUTPUT: {}".format(out_names))
        print("INPUT: {}".format(inp_names))
        model.summary()
        K.set_learning_phase(0)
        sess = K.get_session()
        saver = saver_lib.Saver(write_version=saver_pb2.SaverDef.V2)
        checkpoint_path = saver.save(sess, './saved_ckpt', global_step=0, latest_filename='checkpoint_state')
        graph_io.write_graph(sess.graph, '.', './tmp.pb')
        freeze_graph.freeze_graph('./tmp.pb', '',
                                False, checkpoint_path, out_names,
                                "save/restore_all",
                                "save/Const:0",
                                frozen_model_path,
                                False, "")


import cv2
import skimage.io
import numpy as np
import asyncio
from skimage import img_as_ubyte
from skimage.morphology import convex_hull_image

def draw_box(image, boxs, color=(255,0,0), thickness=2):
  # округление координат
  for box in boxs:
      box = np.int0(box)
      cv2.drawContours(image,[box],0,color,thickness)
  return image

def gamma_lut(img, gamma = 0.5):
    lookUpTable = np.empty((1,256), np.uint8)
    for i in range(256):
        lookUpTable[0,i] = np.clip(pow(i / 255.0, gamma) * 255.0, 0, 255)
    res = cv2.LUT(img, lookUpTable)

async def cv_one_img_mask_async(nn):
    masks = np.array(nn["masks"])
    res = []
    masks = np.array(nn["masks"])
    for i in np.arange(masks.shape[2]):
        mask = np.array([[w[i] for w in h] for h in nn["masks"]], dtype=np.uint8)
        chull = np.array(convex_hull_image(mask), dtype=np.uint8)
        gray = skimage.color.gray2rgb(chull) * 255
        res.append(img_as_ubyte(gray))
    return res

async def cv_img_mask_async(nns):
    loop = asyncio.get_event_loop()
    promises = [loop.create_task(cv_one_img_mask_async(nn)) for nn in nns]
    if bool(promises):
        await asyncio.wait(promises)
    res = []
    for promise in promises:
        res += promise.result()
    return res

def cv_img_mask(nns):
    res = []
    for nn in nns:
        masks = np.array(nn["masks"])
        for i in np.arange(masks.shape[2]):
            mask = np.array([[w[i] for w in h] for h in nn["masks"]], dtype=np.uint8)
            chull = np.array(convex_hull_image(mask), dtype=np.uint8)
            gray = skimage.color.gray2rgb(chull) * 255
            res.append(img_as_ubyte(gray))
    return res

def color_splash(image, nns, color=(0, 255, 0), white_balance=200):
    res = []
    gray = skimage.color.gray2rgb(skimage.color.rgb2gray(image)) * white_balance
    for nn in nns:
        if nn["masks"].shape[-1] > 0:
            mask = (np.sum(nn["masks"], -1, keepdims=True) >= 1)
            fulled = np.full(image.shape, color)
            splash = np.where(mask, fulled, gray).astype(np.uint8)
        else:
            splash = gray.astype(np.uint8)
        res.append(splash)
    return res

def calc_normalize(hist, reverse=0, min_n = 5):
    level = 0
    iterable = hist
    if reverse:
        iterable = reversed(hist)
    for h in iterable:
        if min_n < h:
            break
        level += 1
    if reverse:
        level = len(hist)-level
    return level

def normalize(img, max_p):
    cv_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    hist,bins = np.histogram(cv_img.ravel(),256,[0,255])
    alpha = calc_normalize(hist)
    beta = calc_normalize(hist, reverse=1)
    res = cv2.normalize(cv_img, None, alpha=alpha, beta=beta, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F)


import importlib
import pkgutil

__path__ = pkgutil.extend_path(__path__, __name__)
for importer, modname, ispkg in pkgutil.walk_packages(path=__path__, prefix=__name__+'.'):
    importlib.import_module(modname)


from .xx_xx import xx_xx
import string

class ge(xx_xx):
    def __init__(self):
        super().__init__()
        self.ALLOWED_LITERS = [x for x in string.ascii_letters]
        self.ALLOWED_LITERS.append("0")

        self.STANDARTS = ["@@@###", "@@###@@"]
        self.STANDART = ""


    def find(self, text, strong=False):
        for standart in self.STANDARTS:
            self.STANDART = standart
            match = self.findFully(text)
            if match :
                text = match.group(0)
                newtext = ""
                for i, standart_letter in enumerate(standart):
                    if standart_letter[0] == "@" and text[i] == "0":
                        newtext += "O"
                    else:
                        newtext += text[i]

                return newtext
        return text






import os
import json
import skimage
import numpy as np

from mrcnn.config import Config
from mrcnn import utils

############################################################
#  Configurations
############################################################
class InferenceConfig(Config):
    def __init__(self, config):
         """
             Configuration for training on the  dataset.
             Derives from the base Config class and overrides some values.
             https://github.com/matterport/Mask_RCNN/blob/master/mrcnn/config.py.
         """
         self.__dict__ = config
         super(InferenceConfig, self).__init__()

############################################################
#  Dataset
############################################################
class Dataset(utils.Dataset):
    def load_numberplate(self, subset, config):
        """
                Load a subset of the Numberplate dataset.
                dataset_dir: Root directory of the dataset.
                subset: Subset to load: train or val
        """
        # Add classes. We have only one class to add.
        assert config.CLASS_NAMES and type(config.CLASS_NAMES) == list
        assert config.NAME and type("") == str
        i = 0
        for name in config.CLASS_NAMES[1:]:
            self.add_class(config.NAME, 1, name)
            i += 1

        # Train or validation dataset?
        assert subset in ["train", "val"]
        assert config.DATASET_DIR and type(config.DATASET_DIR) == str
        dataset_dir = os.path.join(config.DATASET_DIR, subset)

        # Load annotations
        # VGG Image Annotator (up to version 1.6) saves each image in the form:
        # { 'filename': '28503151_5b5b7ec140_b.jpg',
        #   'regions': {
        #       '0': {
        #           'region_attributes': {},
        #           'shape_attributes': {
        #               'all_points_x': [...],
        #               'all_points_y': [...],
        #               'name': 'polygon'}},
        #       ... more regions ...
        #   },
        #   'size': 100202
        # }
        # We mostly care about the x and y coordinates of each region
        # Note: In VIA 2.0, regions was changed from a dict to a list.
        annotations = json.load(open(os.path.join(dataset_dir, "via_region_data.json")))
        annotations = annotations['_via_img_metadata']
        annotations = annotations.values()
        if type(annotations) == dict:
            annotations = list(annotations)

        # The VIA tool saves images in the JSON even if they don't have any
        # annotations. Skip unannotated images.
        # print(annotations)
        annotations = [a for a in annotations if a['regions']]

        # Add images
        for a in annotations:
            # Get the x, y coordinaets of points of the polygons that make up
            # the outline of each object instance. These are stores in the
            # shape_attributes (see json format above)
            # The if condition is needed to support VIA versions 1.x and 2.x.
            if type(a['regions']) is dict:
                polygons = [r['shape_attributes'] for r in a['regions'].values()]
            else:
                polygons = [r['shape_attributes'] for r in a['regions']]

            # load_mask() needs the image size to convert polygons to masks.
            # Unfortunately, VIA doesn't include it in JSON, so we must read
            # the image. This is only managable since the dataset is tiny.
            image_path = os.path.join(dataset_dir, a['filename'])
            image = skimage.io.imread(image_path)
            height, width = image.shape[:2]

            self.add_image(
                "numberplate",
                image_id=a['filename'],  # use file name as a unique image id
                path=image_path,
                width=width, height=height,
                polygons=polygons)

    def load_mask(self, image_id):
        """Generate instance masks for an image.
            Returns:
                masks: A bool array of shape [height, width, instance count] with
                one mask per instance.
                class_ids: a 1D array of class IDs of the instance masks.
        """
        # If not a numberplate dataset image, delegate to parent class.
        image_info = self.image_info[image_id]
        if image_info["source"] != "numberplate":
            return super(self.__class__, self).load_mask(image_id)

        # Convert polygons to a bitmap mask of shape
        # [height, width, instance_count]
        info = self.image_info[image_id]
        mask = np.zeros([info["height"], info["width"], len(info["polygons"])],
                        dtype=np.uint8)
        for i, p in enumerate(info["polygons"]):
            # Get indexes of pixels inside the polygon and set them to 1
            rr, cc = skimage.draw.polygon(p['all_points_y'], p['all_points_x'])
            mask[rr, cc, i] = 1

        # Return mask, and array of class IDs of each instance. Since we have
        # one class ID only, we return an array of 1s
        return mask.astype(np.bool), np.ones([mask.shape[-1]], dtype=np.int32)

    def image_reference(self, image_id):
        """Return the path of the image."""
        info = self.image_info[image_id]
        if info["source"] == "numberplate":
            return info["path"]
        else:
            super(self.__class__, self).image_reference(image_id)



import sys, os
import numpy as np
sys.path.append(os.path.abspath(os.path.dirname(__file__)))
import TextDetectors
from tools import np_split

from .mcm.mcm import download_latest_model

class TextDetector():
    @classmethod
    def get_classname(cls):
        return cls.__name__

    def __init__(self, prisets={}):
        self.detectors_map = {}
        self.detectors = []
        self.detectors_names = []

        self.DEFAULT_LABEL = "eu_ua_2015"
        self.DEFAULT_LINES_COUNT = 1

        i = 0
        for prisetName in prisets:
            priset = prisets[prisetName]
            for region in priset["for_regions"]:
                self.detectors_map[region] = i
            _label = prisetName
            if _label not in dir(TextDetectors):
                raise Exception("Text detector {} not in Text Detectors".format(_label))
            TextPostprocessing = getattr(getattr(TextDetectors, _label), _label)
            detector = TextPostprocessing()

            if priset['model_path'].split(".")[-1] == "pb":
                detector.load_frozen(priset['model_path'])
                detector.predict = detector.frozen_predict
            else:
                detector.load(priset['model_path'])
            self.detectors.append(detector)
            self.detectors_names.append(_label)
            i += 1

    def get_avalible_module():
        pass

    def predict(self, zones, labels=None, lines=None, frozen=False):
        if labels is None:
            labels = []
        if lines is None:
            lines = []

        while len(labels) < len(zones):
            labels.append(self.DEFAULT_LABEL)
        while len(lines) < len(zones):
            lines.append(self.DEFAULT_LINES_COUNT)

        zones = np_split(zones, lines)
        predicted = {}

        orderAll = []
        resAll = []
        i = 0
        for zone, label in zip(zones, labels):
            if label in self.detectors_map.keys():
                detector = self.detectors_map[label]
                if detector not in predicted.keys():
                    predicted[detector] = {"zones": [], "order": []}
                predicted[detector]["zones"].append(zone)
                predicted[detector]["order"].append(i)
            else:
                resAll.append("")
                orderAll.append(i)
            i += 1

        for key in predicted.keys():
            if not frozen:
                resAll = resAll + self.detectors[int(key)].predict(predicted[key]["zones"])
            orderAll = orderAll + predicted[key]["order"]

        return [x for _, x in sorted(zip(orderAll,resAll), key=lambda pair: pair[0])]

    @staticmethod
    def get_static_module(name):
        return getattr(getattr(TextDetectors, name), name)

    def get_module(self, name):
        ind = self.detectors_names.index(name)
        return self.detectors[ind]



import numpy
import cv2

def np_split(imgs, lines, coef=0.1):
    res_imgs = []
    for img, line in zip(imgs, lines):
        if line < 2:
            res_imgs.append(img)
        else:
            n = int(img.shape[0]/line + (img.shape[0] * coef))
            firstPart = img[:n]
            for l in range(1, line - 1):
                middlePart = img[:l*n]
                firstPart  = numpy.concatenate((firstPart, middlePart), axis=1)
            lastPart  = img[-n:]
            res_imgs.append(numpy.concatenate((firstPart, lastPart), axis=1))
    return imgs


from os.path import join
import cv2
import os
import json
import numpy as np
from tensorflow.keras import backend as K
import random
import itertools
from .aug import aug

class TextImageGenerator:
    def __init__(self,
                 dirpath,
                 img_w, img_h,
                 batch_size,
                 downsample_factor,
                 letters,
                 max_text_len,
                 cname=""):

        self.CNAME = cname
        #print(self.CNAME)
        self.img_h = img_h
        self.img_w = img_w
        self.batch_size = batch_size
        self.max_text_len = max_text_len
        self.downsample_factor = downsample_factor
        self.letters = letters

        img_dirpath = join(dirpath, 'img')
        ann_dirpath = join(dirpath, 'ann')
        self.samples = []
        for filename in os.listdir(img_dirpath):
            name, ext = os.path.splitext(filename)
            if ext == '.png':
                img_filepath = join(img_dirpath, filename)
                json_filepath = join(ann_dirpath, name + '.json')
                description = json.load(open(json_filepath, 'r'))['description']
                if TextImageGenerator.is_valid_str(self, description):
                    self.samples.append([img_filepath, description])

        self.n = len(self.samples)
        self.indexes = list(range(self.n))
        self.cur_index = 0
        self.count_ep = 0
        self.letters_max = len(letters)+1

    def labels_to_text(self, labels):
        data = ''.join(list(map(lambda x: "" if x==self.letters_max else self.letters[int(x)], labels)))
        return data

    def text_to_labels(self, text):
        data = list(map(lambda x: self.letters.index(x), text))
        while len(data) < self.max_text_len:
            data.append(self.letters_max)
        return data

    def is_valid_str(self, s):
        for ch in s:
            if not ch in self.letters:
                return False
        return True

    def decode_batch(self, out):
        # Most change
        # For a real OCR application, this should be beam search with a dictionary
        # and language model.  For this example, best path is sufficient.
        ret = []
        for j in range(out.shape[0]):
            out_best = list(np.argmax(out[j, 2:], 1))
            out_best = [k for k, g in itertools.groupby(out_best)]
            outstr = ''
            for c in out_best:
                if c < len(self.letters):
                    outstr += self.letters[c]
            ret.append(outstr)
        return ret

    def build_data(self, aug_count=0):
        self.imgs = np.zeros((self.n, self.img_h, self.img_w))
        self.texts = []
        for i, (img_filepath, text) in enumerate(self.samples):
            img = cv2.imread(img_filepath)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

            # CLAHE
            clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8,8))
            img = clahe.apply(img)

            img = cv2.resize(img, (self.img_w, self.img_h))
            img = img.astype(np.float32)
            img -= np.amin(img)
            img /= np.amax(img)
            # width and height are backwards from typical Keras convention
            # because width is the time dimension when it gets fed into the RNN
            self.imgs[i, :, :] = img
            self.texts.append(text)
        while aug_count:
            for i, (img_filepath, text) in enumerate(self.samples):
                img = cv2.imread(img_filepath)
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                imgs = aug([img])
                img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
                img = imgs[0]
                img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

                # CLAHE
                clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8,8))
                img = clahe.apply(img)

                img = cv2.resize(img, (self.img_w, self.img_h))
                img = img.astype(np.float32)
                img -= np.amin(img)
                img /= np.amax(img)
                # width and height are backwards from typical Keras convention
                # because width is the time dimension when it gets fed into the RNN
                self.imgs[i, :, :] = img
                self.texts.append(text)
            aug_count -= 1


    def get_output_size(self):
        return len(self.letters) + 1

    def normalize(self, img):
        img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
        img = cv2.resize(img, (self.IMG_W, self.IMG_H))

        # CLAHE
        clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8,8))
        img = clahe.apply(img)

        img = img.astype(np.float32)
        img -= np.amin(img)
        img /= np.amax(img)
        img = [[[h] for h in w] for w in img.T]

        x = np.zeros((self.IMG_W, self.IMG_H, 1))
        x[:, :, :] = img
        return x

    def next_sample(self, is_random=1):
        self.cur_index += 1
        if self.cur_index >= self.n:
            self.count_ep += 1
            self.cur_index = 0
            if is_random:
                random.shuffle(self.indexes)
        return self.imgs[self.indexes[self.cur_index]], self.texts[self.indexes[self.cur_index]]

    def next_batch(self, is_random=1):
        while True:
            # width and height are backwards from typical Keras convention
            # because width is the time dimension when it gets fed into the RNN
            if K.image_data_format() == 'channels_first':
                X_data = np.ones([self.batch_size, 1, self.img_w, self.img_h])
            else:
                X_data = np.ones([self.batch_size, self.img_w, self.img_h, 1])

            Y_data = np.ones([self.batch_size, self.max_text_len])
            input_length = np.ones((self.batch_size, 1)) * (self.img_w // self.downsample_factor - 2)
            label_length = np.zeros((self.batch_size, 1))
            source_str = []

            for i in range(self.batch_size):
                img, text = self.next_sample(is_random)
                img = img.T
                if K.image_data_format() == 'channels_first':
                    img = np.expand_dims(img, 0)
                else:
                    img = np.expand_dims(img, -1)
                X_data[i] = img
                Y_data[i] = self.text_to_labels(text)
                source_str.append(text)
                label_length[i] = len(text)

            inputs = {
                'the_input_{}'.format(self.CNAME): X_data,
                'the_labels_{}'.format(self.CNAME): Y_data,
                'input_length_{}'.format(self.CNAME): input_length,
                'label_length_{}'.format(self.CNAME): label_length,
                #'source_str': source_str
            }
            outputs = {'ctc': np.zeros([self.batch_size])}
            yield (inputs, outputs)


import sys, os
import asyncio

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
import TextPostprocessings

async def textPostprocessingOneAsync(text, textPostprocessName):
    _textPostprocessName = textPostprocessName.replace("-", "_")
    if _textPostprocessName in dir(TextPostprocessings):
        TextPostprocessing = getattr(getattr(TextPostprocessings, _textPostprocessName), _textPostprocessName)
    else:
        TextPostprocessing = getattr(getattr(TextPostprocessings, "xx_xx"), "xx_xx")
    postprocessManager = TextPostprocessing()
    return postprocessManager.find(text)

async def textPostprocessingAsync(texts, textPostprocessNames):
    loop = asyncio.get_event_loop()
    promises = [loop.create_task(textPostprocessingOneAsync(text, textPostprocessName)) for text, textPostprocessName in zip(texts, textPostprocessNames)]
    if bool(promises):
        await asyncio.wait(promises)
    return [promise.result() for promise in promises]

def textPostprocessing(texts, textPostprocessNames):
    resTexts = []
    for text, textPostprocessName in zip(texts, textPostprocessNames):
        _textPostprocessName = textPostprocessName.replace("-", "_")
        if _textPostprocessName in dir(TextPostprocessings):
            TextPostprocessing = getattr(getattr(TextPostprocessings, _textPostprocessName), _textPostprocessName)
        else:
            TextPostprocessing = getattr(getattr(TextPostprocessings, "xx_xx"), "xx_xx")
        postprocessManager = TextPostprocessing()
        resTexts.append(postprocessManager.find(text))
    return resTexts



import cv2
import keras
import os
import numpy as np
from keras.layers import merge
from keras.optimizers import Adam
from keras.regularizers import l2
from keras.layers.normalization import BatchNormalization
from keras import models
from keras import layers
from keras import backend as K
from keras.models import Model, Input
from keras.preprocessing import image
from keras.applications import VGG16
from keras import callbacks
from keras.models import load_model
import tensorflow as tf
from tensorflow.python.framework import graph_io
from tensorflow.python.tools import freeze_graph
from tensorflow.core.protobuf import saver_pb2
from tensorflow.python.training import saver as saver_lib
from tensorflow.python.framework.graph_util import convert_variables_to_constants

from .mcm.mcm import download_latest_model
from .Base.ImgGenerator import ImgGenerator

class OptionsDetector(ImgGenerator):
    def __init__(self, options = {}):
        # input
        self.DEPTH          = 1
        self.HEIGHT         = 64
        self.WEIGHT         = 295
        self.COLOR_CHANNELS = 3

        # outputs 1
        self.CLASS_REGION = options.get("class_region", ["xx_unknown", "eu_ua_2015", "eu_ua_2004", "eu_ua_1995", "eu", "xx_transit", "ru", "kz"])

        # outputs 2
        self.CLASS_STATE = options.get("class_state", ["BACKGROUND", "FILLED", "NOT_FILLED"])

        # outputs 3
        self.CLASS_COUNT_LINE = options.get("class_count_line", ["0", "1", "2"])

        # model
        self.MODEL = None

        # model hyperparameters
        self.OUT_DENSE_INIT                = 'uniform'
        self.OUT_DENSE_ACTIVATION          = 'softmax'
        self.DENSE_ACTIVATION              = 'softmax'
        self.DROPOUT_1                     = 0.2
        self.DROPOUT_2                     = 0.5
        self.DENSE_LAYERS                  = 512
        self.ENSEMBLES                     = 1
        self.BATCH_NORMALIZATION_AXIS      = -1
        self.L2_LAMBDA                     = 0.001
        self.W_REGULARIZER                 = l2(self.L2_LAMBDA)

        # callbacks hyperparameters
        self.REDUCE_LRO_N_PLATEAU_PATIENCE = 10
        self.REDUCE_LRO_N_PLATEAU_FACTOR   = 0.1

        # train hyperparameters
        self.BATCH_SIZE       = 32
        self.STEPS_PER_EPOCH  = 0 # defain auto
        self.VALIDATION_STEPS = 0 # defain auto
        self.EPOCHS           = 150

        # compile model hyperparameters
        self.LOSSES = {
            "REGION":     "categorical_crossentropy",  # tf.losses.softmax_cross_entropy
            "STATE":      "categorical_crossentropy",  # tf.losses.softmax_cross_entropy
            "COUNT_LINE": "categorical_crossentropy",  # tf.losses.softmax_cross_entropy
        }
        self.LOSS_WEIGHTS = {"REGION": 1.0, "STATE": 1.0, "COUNT_LINE": 1.0}
        self.OPT = "adamax" #  tf.keras.optimizers.Adamax
        self.METRICS = ["accuracy"]

        # for frozen graph
        self.INPUT_NODE = "input_2:0"
        self.OUTPUT_NODES = ("REGION/Softmax:0", "STATE/Softmax:0")

    def change_dimension(self, w, h):
        if w != self.WEIGHT and h != self.HEIGHT:
            self.HEIGHT         = h
            self.WEIGHT         = w
            if self.MODEL != None:
                self.MODEL.layers.pop(0)
                newInput = Input(shape=(self.HEIGHT, self.WEIGHT, self.COLOR_CHANNELS))
                newOutputs = self.MODEL(newInput)
                self.MODEL = Model(newInput, newOutputs)

    def create_model(self, input_model, conv_base, dropout_1, dropout_2, dense_layers, output_labels1, \
                     output_labels2, output_labels3, out_dense_init, W_regularizer, \
                     out_dense_activation, dense_activation, BatchNormalization_axis):
        # cnn
        x = conv_base

        # classificator 1
        x1 = layers.Flatten()(x)
        x1 = layers.Dropout(dropout_2)(x1)
        x1 = layers.Dense(dense_layers, activation=dense_activation)(x1)
        x1 = layers.BatchNormalization(axis=BatchNormalization_axis)(x1)
        x1 = layers.Dense(output_labels1, init=out_dense_init, W_regularizer=W_regularizer)(x1)
        x1 = layers.Activation(out_dense_activation, name="REGION")(x1)

        # classificator 2
        x2 = layers.Flatten()(x)
        x2 = layers.Dropout(dropout_2)(x2)
        x2 = layers.Dense(dense_layers, activation=dense_activation)(x2)
        x2 = layers.BatchNormalization(axis=BatchNormalization_axis)(x2)
        x2 = layers.Dense(output_labels2, init=out_dense_init, W_regularizer=W_regularizer)(x2)
        x2 = layers.Activation(out_dense_activation, name="STATE")(x2)

        # classificator 3
        x3 = layers.Flatten()(x)
        x3 = layers.Dropout(dropout_2)(x3)
        x3 = layers.Dense(dense_layers, activation=dense_activation)(x3)
        x3 = layers.BatchNormalization(axis=BatchNormalization_axis)(x3)
        x3 = layers.Dense(output_labels3, init=out_dense_init, W_regularizer=W_regularizer)(x3)
        x3 = layers.Activation(out_dense_activation, name="COUNT_LINE")(x3)

        #x = keras.layers.concatenate([x1, x2], axis=1)
        model = Model(input=input_model, outputs=[x1, x2, x3])

        # compile model
        model.compile(
            optimizer=self.OPT,
            loss=self.LOSSES,
            loss_weights=self.LOSS_WEIGHTS,
            metrics=self.METRICS
        )
        return model

    def ensemble(self, models, model_input):
        # collect outputs of models in a list
        outputs = [model(model_input) for model in models]

        # averaging outputs
        y = layers.Average()(outputs)

        # build model from same input and avg output
        model = Model(inputs=model_input, outputs=y, name='ensemble')

        return model

    def prepare(self, base_dir, verbose=1):
        if verbose:
            print("START PREPARING")
        # you mast split your data on 3 directory
        train_dir = os.path.join(base_dir, 'train')
        validation_dir = os.path.join(base_dir, 'val')
        test_dir = os.path.join(base_dir, 'test')

        # compile generators
        self.train_generator = self.compile_train_generator(train_dir, (self.HEIGHT, self.WEIGHT), self.BATCH_SIZE)
        self.validation_generator = self.compile_test_generator(validation_dir, (self.HEIGHT, self.WEIGHT), self.BATCH_SIZE)

        self.test_generator = self.compile_test_generator(test_dir, (self.HEIGHT, self.WEIGHT), self.BATCH_SIZE)
        if verbose:
            print("DATA PREPARED")

    def create_conv(self, inp):
         # trainable cnn model
        conv_base = VGG16(weights='imagenet',
            include_top=False)
        # block trainable cnn parameters
        conv_base.trainable = False
        return conv_base(inp)

    def create_simple_conv(self, inp):
        conv_base = layers.Conv2D(32, (3, 3), activation='relu')(inp)
        conv_base = layers.MaxPooling2D((2, 2))(conv_base)

        conv_base = layers.Conv2D(64, (3, 3), activation='relu')(conv_base)
        conv_base = layers.MaxPooling2D((2, 2))(conv_base)

        conv_base = layers.Conv2D(128, (3, 3), activation='relu')(conv_base)
        conv_base = layers.MaxPooling2D((2, 2))(conv_base)

        conv_base = layers.Conv2D(128, (3, 3), activation='relu')(conv_base)
        conv_base = layers.MaxPooling2D((2, 2))(conv_base)

        return conv_base

    def train(self, log_dir="./", verbose=1, cnn="simple"):
        # init count outputs
        self.OTPUT_LABELS_1 = len(self.CLASS_REGION)
        self.OTPUT_LABELS_2 = len(self.CLASS_STATE)
        self.OTPUT_LABELS_3 = len(self.CLASS_COUNT_LINE)

        # create input
        input_model = Input(shape=(self.HEIGHT, self.WEIGHT, self.COLOR_CHANNELS))

        if (cnn == "simple"):
            conv_base = self.create_simple_conv(input_model)
        else:
            conv_base = self.create_conv(input_model)

        # traine callbacks
        self.CALLBACKS_LIST = [
            callbacks.ModelCheckpoint(
                filepath=os.path.join(log_dir, 'buff_weights.h5'),
                monitor='val_loss',
                save_best_only=True,
            ),
            callbacks.ReduceLROnPlateau(
                monitor='val_loss',
                factor=self.REDUCE_LRO_N_PLATEAU_FACTOR,
                patience=self.REDUCE_LRO_N_PLATEAU_PATIENCE,
            )
        ]

        # train all
        modelsArr = []
        for i in np.arange(self.ENSEMBLES):
            # create model
            model = self.create_model(input_model = input_model, conv_base = conv_base, \
                                 dropout_1 = self.DROPOUT_1 , dropout_2 = self.DROPOUT_2, dense_layers = self.DENSE_LAYERS, \
                                 output_labels1 = self.OTPUT_LABELS_1, output_labels2 = self.OTPUT_LABELS_2, \
                                 output_labels3 = self.OTPUT_LABELS_3,
                                 out_dense_init = self.OUT_DENSE_INIT, W_regularizer = self.W_REGULARIZER, \
                                 out_dense_activation = self.OUT_DENSE_ACTIVATION , dense_activation = self.DENSE_ACTIVATION, \
                                 BatchNormalization_axis = self.BATCH_NORMALIZATION_AXIS)

            # train
            history = model.fit_generator(
                self.train_generator,
                steps_per_epoch=self.STEPS_PER_EPOCH,
                epochs=self.EPOCHS,
                callbacks=self.CALLBACKS_LIST,
                validation_data=self.validation_generator,
                validation_steps=self.VALIDATION_STEPS,
                verbose=verbose
            )

            # load best model
            model.load_weights(os.path.join(log_dir, 'buff_weights.h5'))

            # append to models
            modelsArr.append(model)

        # merge ensembles
        if len(modelsArr) > 1:
            self.MODEL = self.ensemble(modelsArr, input_model)
        elif len(modelsArr) == 1:
            self.MODEL = modelsArr[0]

        return  self.MODEL

    def test(self):
        test_loss, test_loss1, test_loss2, test_loss3, test_acc1, test_acc2, test_acc3 = self.MODEL.evaluate_generator(self.test_generator, steps=self.VALIDATION_STEPS)
        print("test loss: {}".format(test_loss))
        print("test loss: {}    test loss: {}       test loss: {}".format(test_loss1, test_loss2, test_loss3))
        print("test acc: {}    test acc {}      test acc {}".format(test_acc1, test_acc2, test_acc3))
        return test_loss, test_loss1, test_loss2, test_loss3, test_acc1, test_acc2, test_acc3

    def save(self, path, verbose=1):
        if self.MODEL != None:
            if bool(verbose):
                print("model save to {}".format(path))
            self.MODEL.save(path)

    def isLoaded(self):
        if self.MODEL == None:
            return False
        return True

    @classmethod
    def get_classname(cls):
        return cls.__name__

    def load(self, path_to_model, options={}, verbose = 0):
        if path_to_model == "latest":
            model_info   = download_latest_model(self.get_classname(), "simple")
            path_to_model   = model_info["path"]
            options["class_region"] = model_info["class_region"]

        self.CLASS_REGION = options.get("class_region", ["xx_unknown", "eu_ua_2015", "eu_ua_2004", "eu_ua_1995", "eu", "xx_transit", "ru", "kz"])
        if path_to_model.split(".")[-1] != "pb":
            self.MODEL = load_model(path_to_model)
            if verbose:
                self.MODEL.summary()
        else:
            self.load_frozen(path_to_model)
            self.predict = self.frozen_predict

    def getRegionLabel(self, index):
        return self.CLASS_REGION[index]

    def getStateLabel(self, index):
        return self.CLASS_STATE[index]

    def predict(self, imgs):
        Xs = []
        for img in imgs:
            Xs.append(self.normalize(img))

        predicted = [[], [], []]
        if bool(Xs):
            predicted = self.MODEL.predict(np.array(Xs))
	    
        regionIds = []
        for region in predicted[0]:
            regionIds.append(int(np.argmax(region)))

        stateIds = []
        for state in predicted[1]:
            stateIds.append(int(np.argmax(state)))

        countLines = []
        for countL in predicted[2]:
            countLines.append(int(np.argmax(countL)))

        return regionIds, stateIds, countLines

    def getRegionLabels(self, indexes):
        return [self.CLASS_REGION[index] for index in indexes]

    def compile_train_generator(self, train_dir, target_size, batch_size=32):
        # with data augumentation
        imageGenerator = ImgGenerator(train_dir, self.WEIGHT, self.HEIGHT, self.BATCH_SIZE, [len(self.CLASS_STATE), len(self.CLASS_REGION), len(self.CLASS_COUNT_LINE)])
        print("start train build")
        imageGenerator.build_data()
        self.STEPS_PER_EPOCH = self.STEPS_PER_EPOCH or imageGenerator.n / imageGenerator.batch_size or imageGenerator.n / imageGenerator.batch_size + 1
        print("end train build")
        return  imageGenerator.generator()

    def compile_test_generator(self, test_dir, target_size, batch_size=32):
        imageGenerator = ImgGenerator(test_dir, self.WEIGHT, self.HEIGHT, self.BATCH_SIZE, [len(self.CLASS_STATE), len(self.CLASS_REGION), len(self.CLASS_COUNT_LINE)])
        print("start test build")
        imageGenerator.build_data()
        self.VALIDATION_STEPS = self.VALIDATION_STEPS or imageGenerator.n / imageGenerator.batch_size or imageGenerator.n / imageGenerator.batch_size + 1
        print("end test build")
        return  imageGenerator.generator()

    def load_frozen(self, FROZEN_MODEL_PATH):
        graph_def = tf.GraphDef()
        with tf.gfile.GFile(FROZEN_MODEL_PATH, "rb") as f:
            graph_def.ParseFromString(f.read())

        graph = tf.Graph()
        with graph.as_default():
            self.net_inp, self.net_out1, self.net_out2 = tf.import_graph_def(
                graph_def, return_elements = [self.INPUT_NODE, self.OUTPUT_NODES[0], self.OUTPUT_NODES[1]]
            )
            #print([x.name for x in graph_def.node])

        sess_config = tf.ConfigProto()
        sess_config.gpu_options.allow_growth = True
        self.sess = tf.Session(graph=graph, config=sess_config)

    def frozen_predict(self, imgs):
        Xs = []
        for img in imgs:
            img = self.normalize(img)
            Xs.append(img)

        predicted = [[], []]
        if bool(Xs):
            predicted = self.sess.run([self.net_out1, self.net_out2], feed_dict={self.net_inp:Xs})

        regionIds = []
        for region in predicted[0]:
            regionIds.append(int(np.argmax(region)))

        stateIds = []
        for state in predicted[1]:
            stateIds.append(int(np.argmax(state)))

        return regionIds, stateIds


import sys
import matplotlib.image as mpimg
import tensorflow as tf
import numpy as np
from tensorflow.python.framework import graph_io
from tensorflow.python.tools import freeze_graph
from tensorflow.core.protobuf import saver_pb2
from tensorflow.python.training import saver as saver_lib
from tensorflow.python.framework.graph_util import convert_variables_to_constants
import keras

config = tf.ConfigProto()
config.gpu_options.allow_growth = True
session = tf.Session(config=config)

from .mcm.mcm import download_latest_model

class Detector:
    def __init__(self, mask_rcnn_dir, log_dir, mask_rcnn_config = None):
        if mask_rcnn_dir != None:
            self.MASK_RCNN_DIR = mask_rcnn_dir
            self.LOG_DIR = log_dir

            DEFAULT_MASK_RCNN_CONFIG = {
              "NAME": "numberplate",
              "GPU_COUNT": 1,
              "IMAGES_PER_GPU": 1,
              "NUM_CLASSES": 2,
              "DETECTION_MIN_CONFIDENCE": 0.7,
              "IMAGE_MAX_DIM": 1024, # work ?
              "IMAGE_RESIZE_MODE": "square" # work ?
            }
            self.NN_MASK_RCNN_CONFIG = mask_rcnn_config or DEFAULT_MASK_RCNN_CONFIG
            sys.path.append(mask_rcnn_dir)

            from .nnmrcnn import InferenceConfig
            self.CONFIG = InferenceConfig(self.NN_MASK_RCNN_CONFIG)

            # for frozen graph
            self.INPUT_NODES = ("input_image:0", "input_image_meta:0", "input_anchors:0") # 3 elem

            self.OUTPUT_NODES = ("mrcnn_detection/Reshape_1:0", "mrcnn_class/Reshape_1:0", "mrcnn_bbox/Reshape:0", "mrcnn_mask/Reshape_1:0",
                                 "ROI/packed_2:0", "rpn_class/concat:0", "rpn_bbox/concat:0") # 7 elem

             #from mrcnn import visualize
             #self.visualize = visualize


    def loadFrozenModel(self, FROZEN_MODEL_PATH):
        graph_def = tf.GraphDef()
        with tf.gfile.GFile(FROZEN_MODEL_PATH, "rb") as f:
            graph_def.ParseFromString(f.read())

        graph = tf.Graph()
        with graph.as_default():
            self.input_image, self.input_image_meta, self.input_anchors, self.mrcnn_detection, self.mrcnn_class, self.mrcnn_bbox, self.mrcnn_mask, self.ROI, self.rpn_class, self.rpn_bbox = tf.import_graph_def(
                graph_def, return_elements = self.INPUT_NODES + self.OUTPUT_NODES
            )

        sess_config = tf.ConfigProto()
        sess_config.gpu_options.allow_growth = True
        self.sess = tf.Session(graph=graph, config=sess_config)

    @classmethod
    def get_classname(cls):
        return cls.__name__

    def loadModel(self, model_path="latest", verbose = 0, mode="inference"):
        if model_path == "latest":
            model_info = download_latest_model(self.get_classname(), "mrcnn")
            model_path = model_info["path"]

        import mrcnn.model as modellib
        self.MODEL = modellib.MaskRCNN(mode="inference", model_dir=self.LOG_DIR, config=self.CONFIG)
        if model_path.split(".")[-1] == "pb":
            self.loadFrozenModel(model_path)
            self.detect = self.frozen_detect
        else:
            self.MODEL.load_weights(model_path, by_name=True)

    def getKerasModel(self):
        return self.MODEL.keras_model

    def normalize(self, images):
        res = []
        for image in images:
            # delete 4 chanel
            res.append(image[..., :3])
        return res

    def detectFromFile(self, image_paths, verbose = 0):
        images = [mpimg.imread(image_path) for image_path in image_paths]
        return self.detect(images, verbose=verbose)

    def detect(self, images, verbose = 0):
        return self.MODEL.detect(self.normalize(images), verbose=verbose)

#    def detect_masks(self, images, verbose = 0):
#        r = self.MODEL.detect(self.normalize(images), verbose=verbose)
#        # loop over of the detected object's bounding boxes and masks
#        for i in range(0, r["rois"].shape[0]):
#            # extract the class ID and mask for the current detection, then
#            # grab the color to visualize the mask (in BGR format)
#            classID = r["class_ids"][i]
#            mask = r["masks"][:, :, i]
#            color = COLORS[classID][::-1]
#
#            # visualize the pixel-wise mask of the object
#            image = visualize.apply_mask(image, mask, color, alpha=0.5)



    def frozen_detect(self, images, verbose = 0):
        if verbose:
            print(self.CONFIG.BATCH_SIZE)
        assert len(images) == self.CONFIG.BATCH_SIZE, "len(images) must be equal to BATCH_SIZE"

        # Mold inputs to format expected by the neural network
        molded_images, image_metas, windows = self.MODEL.mold_inputs(images)

        # Validate image sizes
        # All images in a batch MUST be of the same size
        image_shape = molded_images[0].shape
        for g in molded_images[1:]:
            assert g.shape == image_shape,\
                "After resizing, all images must have the same size. Check IMAGE_RESIZE_MODE and image sizes."

        # Anchors
        anchors = self.MODEL.get_anchors(image_shape)
        anchors = np.broadcast_to(anchors, (self.CONFIG.BATCH_SIZE,) + anchors.shape)

        if verbose:
            print("molded_images", molded_images)
            print("image_metas", image_metas)
            print("anchors", anchors)

        #print("Run detection")
        # Run object detection
        detections, _, _, mrcnn_mask, _, _, _ = self.sess.run([self.mrcnn_detection, self.mrcnn_class, self.mrcnn_bbox, self.mrcnn_mask, self.ROI, self.rpn_class, self.rpn_bbox], feed_dict={
                    self.input_image:      molded_images,
                    self.input_image_meta: image_metas,
                    self.input_anchors:    anchors
                })

        # Process detections
        results = []
        for i, image in enumerate(images):
            final_rois, final_class_ids, final_scores, final_masks =\
                self.MODEL.unmold_detections(detections[i], mrcnn_mask[i],
                                       image.shape, molded_images[i].shape,
                                       windows[i])
            results.append({
                "rois": final_rois,
                "class_ids": final_class_ids,
                "scores": final_scores,
                "masks": final_masks,
            })
        return results

    ############################################################
    #  Training
    ############################################################
    def train(self, augmentation=None, verbose=1):
        keras.backend.clear_session()
        from .nnmrcnn import Dataset
        import mrcnn.model as modellib
        from mrcnn import utils
        from imgaug import augmenters as iaa
        if verbose:
            self.CONFIG.display()
        model = modellib.MaskRCNN(mode="training", config=self.CONFIG, model_dir=self.LOG_DIR)

        # Select weights file to load
        assert self.CONFIG.WEIGHTS and type(self.CONFIG.WEIGHTS) is str

        if self.CONFIG.WEIGHTS.lower() == "coco":
            weights_path = os.path.join(self.LOG_DIR, "mask_rcnn_coco.h5")
            # Download weights file
            if not os.path.exists(weights_path):
                utils.download_trained_weights(weights_path)
        elif self.CONFIG.WEIGHTS.lower() == "last":
            # Find last trained weights
            weights_path = model.find_last()
        elif self.CONFIG.WEIGHTS.lower() == "imagenet":
            # Start from ImageNet trained weights
            weights_path = model.get_imagenet_weights()
        else:
            weights_path = self.CONFIG.WEIGHTS

        # Load weights
        if verbose:
            print("Loading weights ", weights_path)
        if self.CONFIG.WEIGHTS.lower() == "coco":
            # Exclude the last layers because they require a matching
            # number of classes
            model.load_weights(weights_path, by_name=True, exclude=[
                "mrcnn_class_logits", "mrcnn_bbox_fc", "mrcnn_bbox", "mrcnn_mask"])
        else:
            model.load_weights(weights_path, by_name=True)

        # Training dataset.
        if verbose:
            print("Prepare train data")
        dataset_train = Dataset()
        dataset_train.load_numberplate("train", self.CONFIG)
        dataset_train.prepare()

        # Validation dataset
        if verbose:
            print("Prepare validation data")
        dataset_val = Dataset()
        dataset_val.load_numberplate("val", self.CONFIG)
        dataset_val.prepare()

        # Image augmentation
        # http://imgaug.readthedocs.io/en/latest/source/augmenters.html
        augmentation_default = iaa.SomeOf((0, 2), [
            iaa.Fliplr(0.5),
            iaa.Flipud(0.5),
            iaa.OneOf([iaa.Affine(rotate=90),
                       iaa.Affine(rotate=180),
                       iaa.Affine(rotate=270)]),
            iaa.Multiply((0.8, 1.5)),
            iaa.GaussianBlur(sigma=(0.0, 5.0))
        ])
        augmentation = augmentation or augmentation_default

        # *** This training schedule is an example. Update to your needs ***
        # Since we're using a very small dataset, and starting from
        # COCO trained weights, we don't need to train too long. Also,
        # no need to train all layers, just the heads should do it.
        if verbose:
            print("Training network")
        model.train(dataset_train, dataset_val,
                    learning_rate=self.CONFIG.LEARNING_RATE,
                    epochs=self.CONFIG.EPOCHS,
                    augmentation=augmentation,
                    layers=self.CONFIG.LAYERS )



from setuptools import setup, find_packages

setup(name='nomeroff-net',
      version='0.3.3',
      description='Automatic numberplate recognition system',
      long_description='Nomeroff Net is a opensource python license plate recognition framework based on the application of a convolutional neural network on the Mask_RCNN architecture, and cusomized OCR-module powered by GRU architecture. The project is now at the initial stage of development, write to us if you are interested in helping us in the formation of a dataset for your country.',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
      ],
      keywords='Nomeroff Net ai  Mask_RCNN OCR GRU opensource license number plate recognition licenseplate numberplate license-plate number-plate ria-com RIA.com ria',
      url='https://github.com/ria-com/nomeroff-net',
      author='Dmytro Probachay, Oleg Cherniy',
      author_email='dimabendera@gmail.com, oleg.cherniy@ria.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'numpy',
          'matplotlib',
          'Keras',
          'Cython',
          'pycocotools',
          'matplotlib',
          'tensorflow',
          'opencv_python',
          'imutils',
          'scikit_image',
          'jupyter',
          'asyncio',
          'imgaug'
      ],
      dependency_links=[
        'https://github.com/aleju/imgaug',
        'https://github.com/matterport/Mask_RCNN'
      ],
      include_package_data=True,
      zip_safe=False)


# import labaris
import tensorflow.keras
import tensorflow as tf

import os
from os.path import join
import json
import random
import numpy as np

from keras import backend as K
from keras.layers.convolutional import Conv2D, MaxPooling2D
from keras.layers import Input, Dense, Activation
from keras.layers import Reshape, Lambda
from keras.layers.merge import add, concatenate
from keras.models import Model, load_model
from keras.optimizers import SGD
from keras.utils.data_utils import get_file
from keras.preprocessing import image
import keras.callbacks
from collections import Counter

from keras.layers import CuDNNGRU as GRUgpu
from keras.layers.recurrent import GRU as GRUcpu

from .TextImageGenerator import TextImageGenerator
from NomeroffNet.mcm.mcm import download_latest_model

config = tf.ConfigProto()
config.gpu_options.allow_growth = True
session = tf.Session(config=config)

class OCR(TextImageGenerator):
    @classmethod
    def get_classname(cls):
        return cls.__name__

    def __init__(self):
        # Input parameters
        self.IMG_H = 64
        self.IMG_W = 128
        self.IMG_C = 1

        # Train parameters
        self.BATCH_SIZE = 32
        self.EPOCHS = 1

        # Network parameters
        self.CONV_FILTERS = 16
        self.KERNEL_SIZE = (3, 3)
        self.POOL_SIZE = 2
        self.TIME_DENSE_SIZE = 32
        self.RNN_SIZE = 512
        self.ACTIVATION = 'relu'
        self.DOWNSAMPLE_FACROT = self.POOL_SIZE * self.POOL_SIZE

        self.GRU = GRUcpu

        self.INPUT_NODE = "the_input_{}:0".format(type(self).__name__)
        self.OUTPUT_NODE = "softmax_{}/truediv:0".format(type(self).__name__)

    def get_counter(self, dirpath, verbose=1):
        dirname = os.path.basename(dirpath)
        ann_dirpath = join(dirpath, 'ann')
        letters = ''
        lens = []
        for filename in os.listdir(ann_dirpath):
            json_filepath = join(ann_dirpath, filename)
            description = json.load(open(json_filepath, 'r'))['description']
            lens.append(len(description))
            letters += description
        max_plate_length = max(Counter(lens).keys())
        if verbose:
            print('Max plate length in "%s":' % dirname, max_plate_length)
        return Counter(letters), max_plate_length

    def get_alphabet(self, train_path, test_path, val_path, verbose=1):
        c_val, max_plate_length_val     = self.get_counter(val_path)
        c_train, max_plate_length_train = self.get_counter(train_path)
        c_test, max_plate_length_test   = self.get_counter(test_path)

        letters_train = set(c_train.keys())
        letters_val = set(c_val.keys())
        letters_test = set(c_test.keys())
        if verbose:
            print("Letters train ", letters_train)
            print("Letters val ", letters_val)
            print("Letters test ", letters_test)

        if max_plate_length_val == max_plate_length_train:
            if verbose:
                print('Max plate length in train, test and val do match')
        else:
            raise Exception('Max plate length in train, test and val do not match')

        if letters_train == letters_val:
            if verbose:
                print('Letters in train, val and test do match')
        else:
            raise Exception('Letters in train, val and test do not match')

        self.letters = sorted(list(letters_train))
        self.max_text_len = max_plate_length_train
        if verbose:
            print('Letters:', ' '.join(self.letters))
        return self.letters, self.max_text_len

    def explainTextGenerator(self, train_dir, letters, max_plate_length, verbose=1):
        tiger = TextImageGenerator(train_dir, self.IMG_W, self.IMG_H, 1, self.POOL_SIZE * self.POOL_SIZE, letters, max_plate_length, cname=type(self).__name__)
        tiger.build_data()

        for inp, out in tiger.next_batch():
            print('Text generator output (data which will be fed into the neutral network):')
            print('1) the_input (image)')
            if K.image_data_format() == 'channels_first':
                img = inp['the_input_{}'.format(type(self).__name__)][0, 0, :, :]
            else:
                img = inp['the_input_{}'.format(type(self).__name__)][0, :, :, 0]

            #import matplotlib.pyplot as plt
            #plt.imshow(img.T, cmap='gray')
            #plt.show()
            print('2) the_labels (plate number): %s is encoded as %s' %
                  (tiger.labels_to_text(inp['the_labels_{}'.format(type(self).__name__)][0]), list(map(int, inp['the_labels_{}'.format(type(self).__name__)][0]))))
            print('3) input_length (width of image that is fed to the loss function): %d == %d / 4 - 2' %
                  (inp['input_length_{}'.format(type(self).__name__)][0], tiger.img_w))
            print('4) label_length (length of plate number): %d' % inp['label_length_{}'.format(type(self).__name__)][0])
            break

    def ctc_lambda_func(self, args):
        y_pred, labels, input_length, label_length = args
        # the 2 is critical here since the first couple outputs of the RNN
        # tend to be garbage:
        y_pred = y_pred[:, 2:, :]
        return K.ctc_batch_cost(labels, y_pred, input_length, label_length)

    def save(self, path, verbose=1):
        if self.MODEL:
            self.MODEL.save(path)
            if verbose:
                print("SAVED TO {}".format(path))

    def test(self, verbose=1):
        if verbose:
            print("\nRUN TEST")
        net_inp = self.MODEL.get_layer(name='the_input_{}'.format(type(self).__name__)).input
        net_out = self.MODEL.get_layer(name='softmax_{}'.format(type(self).__name__)).output

        err_c = 0
        succ_c = 0
        for inp_value, _ in self.tiger_test.next_batch():
            bs = inp_value['the_input_{}'.format(type(self).__name__)].shape[0]
            X_data = inp_value['the_input_{}'.format(type(self).__name__)]
            net_out_value = self.SESS.run(net_out, feed_dict={net_inp:X_data})
            pred_texts = self.tiger_test.decode_batch(net_out_value)
            labels = inp_value['the_labels_{}'.format(type(self).__name__)]
            texts = []
            for label in labels:
                text = self.tiger_test.labels_to_text(label)
                texts.append(text)

            for i in range(bs):
                if (pred_texts[i] != texts[i]):
                    if verbose:
                        print('\nPredicted: \t\t %s\nTrue: \t\t\t %s' % (pred_texts[i], texts[i]))
                    err_c += 1
                else:
                    succ_c += 1
            break
        print("acc: {}".format(succ_c/(err_c+succ_c)))

    def predict(self, imgs, *argv):
        Xs = []
        for img in imgs:
            x = self.normalize(img)
            Xs.append(x)
        pred_texts = []
        if bool(Xs):
            net_out_value = self.MODEL.predict(np.array(Xs))
            #print(net_out_value)
            pred_texts = self.decode_batch(net_out_value)
        return pred_texts

    def load(self, path_to_model, verbose = 0):
        if path_to_model =="latest":
            model_info = download_latest_model("TextDetector", self.get_classname())
            path_to_model = model_info["path"]


        self.MODEL = load_model(path_to_model, compile=False)

        net_inp = self.MODEL.get_layer(name='the_input_{}'.format(type(self).__name__)).input
        net_out = self.MODEL.get_layer(name='softmax_{}'.format(type(self).__name__)).output

        self.MODEL = Model(input=net_inp, output=net_out)

        if verbose:
            self.MODEL.summary()

        # the loss calc occurs elsewhere, so use a dummy lambda func for the loss
        #sgd = SGD(lr=0.02, decay=1e-6, momentum=0.9, nesterov=True, clipnorm=5)
        #self.MODEL.compile(loss={'ctc': lambda y_true, y_pred: y_pred}, optimizer=sgd)

        return self.MODEL

    def prepare(self, path_to_dataset, aug_count=0, verbose=1):
        self.SESS = tf.Session()
        K.set_session(self.SESS)

        train_path = os.path.join(path_to_dataset, "train")
        test_path  = os.path.join(path_to_dataset, "test")
        val_path   = os.path.join(path_to_dataset, "val")

        if verbose:
            print("GET ALPHABET")
        self.letters, max_plate_length = self.get_alphabet(train_path, test_path, val_path, verbose=verbose)

        if verbose:
            print("\nEXPLAIN DATA TRANSFORMATIONS")
            self.explainTextGenerator(train_path, self.letters, max_plate_length)

        if verbose:
            print("START BUILD DATA")
        self.tiger_train = TextImageGenerator(train_path, self.IMG_W, self.IMG_H, self.BATCH_SIZE, self.DOWNSAMPLE_FACROT, self.letters, max_plate_length, cname=type(self).__name__)
        self.tiger_train.build_data(aug_count=aug_count)
        self.tiger_val = TextImageGenerator(val_path,  self.IMG_W, self.IMG_H, self.BATCH_SIZE, self.DOWNSAMPLE_FACROT, self.letters, max_plate_length, cname=type(self).__name__)
        self.tiger_val.build_data()

        self.tiger_test = TextImageGenerator(test_path, self.IMG_W, self.IMG_H, len(os.listdir(os.path.join(test_path, "img"))), self.DOWNSAMPLE_FACROT, self.letters, max_plate_length, cname=type(self).__name__)
        self.tiger_test.build_data()
        if verbose:
            print("DATA PREPARED")

    def train(self, mode="cpu", is_random=1, model_path="./model.h5", load=False, verbose=1):
        if mode == "gpu":
            self.GRU = GRUgpu
        if mode == "cpu":
            self.GRU = GRUcpu

        if verbose:
            print("\nSTART TRAINING")
        if K.image_data_format() == 'channels_first':
            input_shape = (1, self.IMG_W, self.IMG_H)
        else:
            input_shape = (self.IMG_W, self.IMG_H, 1)

        input_data = Input(name='the_input_{}'.format(type(self).__name__), shape=input_shape, dtype='float32')
        inner = Conv2D(self.CONV_FILTERS, self.KERNEL_SIZE, padding='same',
                       activation=self.ACTIVATION, kernel_initializer='he_normal',
                       name='conv1')(input_data)
        inner = MaxPooling2D(pool_size=(self.POOL_SIZE , self.POOL_SIZE ), name='max1')(inner)
        inner = Conv2D(self.CONV_FILTERS, self.KERNEL_SIZE, padding='same',
                       activation=self.ACTIVATION, kernel_initializer='he_normal',
                       name='conv2')(inner)
        inner = MaxPooling2D(pool_size=(self.POOL_SIZE , self.POOL_SIZE ), name='max2')(inner)

        conv_to_rnn_dims = (self.IMG_W // (self.POOL_SIZE  * self.POOL_SIZE), (self.IMG_H // (self.POOL_SIZE * self.POOL_SIZE)) * self.CONV_FILTERS)
        inner = Reshape(target_shape=conv_to_rnn_dims, name='reshape')(inner)

        # cuts down input size going into RNN:
        inner = Dense(self.TIME_DENSE_SIZE, activation=self.ACTIVATION, name='dense1')(inner)

        # Two layers of bidirecitonal GRUs
        # GRU seems to work as well, if not better than LSTM:
        gru_1 = self.GRU(self.RNN_SIZE, return_sequences=True, kernel_initializer='he_normal', name='gru1')(inner)
        gru_1b = self.GRU(self.RNN_SIZE, return_sequences=True, go_backwards=True, kernel_initializer='he_normal', name='gru1_b')(inner)
        gru1_merged = add([gru_1, gru_1b])
        gru_2 = self.GRU(self.RNN_SIZE, return_sequences=True, kernel_initializer='he_normal', name='gru2')(gru1_merged)
        gru_2b = self.GRU(self.RNN_SIZE, return_sequences=True, go_backwards=True, kernel_initializer='he_normal', name='gru2_b')(gru1_merged)

        # transforms RNN output to character activations:
        inner = Dense(self.tiger_train.get_output_size(), kernel_initializer='he_normal',
                      name='dense2')(concatenate([gru_2, gru_2b]))
        y_pred = Activation('softmax', name='softmax_{}'.format(type(self).__name__))(inner)
        Model(inputs=input_data, outputs=y_pred).summary()

        labels = Input(name='the_labels_{}'.format(type(self).__name__), shape=[self.tiger_train.max_text_len], dtype='float32')
        input_length = Input(name='input_length_{}'.format(type(self).__name__), shape=[1], dtype='int64')
        label_length = Input(name='label_length_{}'.format(type(self).__name__), shape=[1], dtype='int64')
        # Keras doesn't currently support loss funcs with extra parameters
        # so CTC loss is implemented in a lambda layer
        loss_out = Lambda(self.ctc_lambda_func, output_shape=(1,), name='ctc')([y_pred, labels, input_length, label_length])

        # clipnorm seems to speeds up convergence
        sgd = SGD(lr=0.02, decay=1e-6, momentum=0.9, nesterov=True, clipnorm=5)

        if load:
            model = load_model(model_path, compile=False)
        else:
            model = Model(inputs=[input_data, labels, input_length, label_length], outputs=loss_out)

        # the loss calc occurs elsewhere, so use a dummy lambda func for the loss
        model.compile(loss={'ctc': lambda y_true, y_pred: y_pred}, optimizer=sgd)

        if not load:
            # captures output of softmax so we can decode the output during visualization
            test_func = K.function([input_data], [y_pred])

            model.fit_generator(generator=self.tiger_train.next_batch(is_random),
                                steps_per_epoch=self.tiger_train.n,
                                epochs=self.EPOCHS,
                                validation_data=self.tiger_val.next_batch(is_random),
                                validation_steps=self.tiger_val.n)

        net_inp = model.get_layer(name='the_input_{}'.format(type(self).__name__)).input
        net_out = model.get_layer(name='softmax_{}'.format(type(self).__name__)).output
        self.MODEL = Model(input=net_inp, output=net_out)
        return self.MODEL

    def load_frozen(self, FROZEN_MODEL_PATH):
        graph_def = tf.GraphDef()
        with tf.gfile.GFile(FROZEN_MODEL_PATH, "rb") as f:
            graph_def.ParseFromString(f.read())

        #print([x.name for x in graph_def.node])
        graph = tf.Graph()
        with graph.as_default():
            self.net_inp, self.net_out = tf.import_graph_def(
                graph_def, return_elements = [self.INPUT_NODE, self.OUTPUT_NODE]
            )

        sess_config = tf.ConfigProto()
        sess_config.gpu_options.allow_growth = True
        self.sess = tf.Session(graph=graph, config=sess_config)

    def frozen_predict(self, imgs):
        Xs = []
        for img in imgs:
            x = self.normalize(img)
            Xs.append(x)
        pred_texts = []
        if bool(Xs):
            net_out_value = self.sess.run([self.net_out], feed_dict={self.net_inp: np.array(Xs)})
            #print(net_out_value)
            pred_texts = self.decode_batch(net_out_value[0])
        return pred_texts